import { ɵglobal, Directive, ElementRef, Input, NgModule, ChangeDetectorRef, NgZone, TemplateRef, ViewContainerRef, Output, Optional, Pipe } from '@angular/core';
import { Subscription, BehaviorSubject, ReplaySubject, of, concat, NEVER, Subject, defer, Observable } from 'rxjs';
import { RxBaseTemplateNames, templateNotifier, hotFlatten, createTemplateManager, RxNotificationKind, RxStrategyProvider, strategyHandling } from '@rx-angular/cdk';
import { materialize, tap, map, distinctUntilChanged, switchMap, shareReplay, filter, withLatestFrom, startWith, publish, mergeAll } from 'rxjs/operators';
import { coerceBooleanProperty } from '@angular/cdk/coercion';

const zonePatchedEvents = [
    'scroll',
    'mousedown',
    'mouseenter',
    'mouseleave',
    'mousemove',
    'mouseout',
    'mouseover',
    'mouseup',
    'load',
    'pointerup',
    'change',
    'blur',
    'focus',
    'click',
    'contextmenu',
    'drag',
    'dragend',
    'dragenter',
    'dragleave',
    'dragover',
    'dragstart',
    'drop',
    'input'
];

/**
 * @description
 *
 * A fallback for the new `globalThis` reference.
 *
 *  It should be used to replace `window` due to different environments in:
 *  - SSR (Server Side Rendering)
 *  - Tests
 *  - Browser
 *
 *  @return {globalThis} - A reference to globalThis. `window` in the Browser.
 */
function getGlobalThis() {
    return ɵglobal;
}

/*
 * createPropertiesWeakMap
 *
 * @param getDefaults: (o: O) => P
 * Example:
 *
 * export interface Properties {
 *   isCoalescing: boolean;
 * }
 *
 * const obj: object = {
 *   foo: 'bar',
 *   isCoalescing: 'weakMap version'
 * };
 *
 * const getDefaults = (ctx: object): Properties => ({isCoalescing: false});
 * const propsMap = createPropertiesWeakMap<object, Properties>(getDefaults);
 *
 * console.log('obj before:', obj);
 * // {foo: "bar", isCoalescing: "weakMap version"}
 * console.log('props before:', propsMap.getProps(obj));
 * // {isCoalescing: "weakMap version"}
 *
 * propsMap.setProps(obj, {isCoalescing: true});
 * console.log('obj after:', obj);
 * // {foo: "bar", isCoalescing: "weakMap version"}
 * console.log('props after:', propsMap.getProps(obj));
 * // {isCoalescing: "true"}
 * */
function createPropertiesWeakMap(getDefaults) {
    const propertyMap = new WeakMap();
    return {
        getProps: getProperties,
        setProps: setProperties
    };
    function getProperties(ctx) {
        const defaults = getDefaults(ctx);
        const propertiesPresent = propertyMap.get(ctx);
        let properties;
        if (propertiesPresent !== undefined) {
            properties = propertiesPresent;
        }
        else {
            properties = {};
            Object.entries(defaults).forEach(([prop, value]) => {
                properties[prop] = hasKey(ctx, prop) ? ctx[prop] : value;
            });
            propertyMap.set(ctx, properties);
        }
        return properties;
    }
    function setProperties(ctx, props) {
        const properties = getProperties(ctx);
        Object.entries(props).forEach(([prop, value]) => {
            properties[prop] = value;
        });
        propertyMap.set(ctx, properties);
        return properties;
    }
    function hasKey(ctx, property) {
        return ctx[property] != null;
    }
}

/**
 * getZoneUnPatchedApi
 *
 * @description
 *
 * This function returns the zone un-patched API for the a specific Browser API.
 * If no element is passed the window is used instead
 *
 * @param name {string} - The name of the API to check.
 * @param elem {any} - The elem to get un-patched API from.
 * @return {Function} - The zone un-patched API in question.
 *
 */
function getZoneUnPatchedApi(name, elem) {
    elem = elem || getGlobalThis();
    return isApiZonePatched(name, elem) ? elem['__zone_symbol__' + name] : elem[name];
}
/**
 * envZonePatched
 *
 * @description
 *
 * This function checks the window object `zone.js` was instantiated.
 * If so, the `window` object maintains a property named `Zone`.
 *
 * Here how Angular checks it: https://github.com/angular/angular/blob/master/packages/core/src/zone/ng_zone.ts#L123
 *
 * @return {boolean} - true if `zone.js` patched global APIs.
 *
 */
function isEnvZonePatched() {
    return getGlobalThis().Zone !== undefined;
}
/**
 * apiZonePatched
 *
 * @description
 *
 * This function checks if a specific Browser API is patched by `zone.js`.
 *
 * @param name - The name of the API to check.
 * @param elem - The name of the API to check.
 * @return {boolean} - true if `zone.js` patched the API in question.
 *
 */
function isApiZonePatched(name, elem) {
    // if symbol is present, zone patched the API
    return elem['__zone_symbol__' + name] !== undefined;
}
const zoneDetectionCache = new WeakMap();
/**
 * isNgZone
 *
 * @description
 *
 * This function takes an instance of a class which implements the NgZone interface and checks if
 * its `runOutsideAngular()` function calls `apply()` on the function passed as parameter. This
 * means the Angular application that instantiated this service assumes it runs in a ZoneLess
 * environment, and therefore it's change detection will not be triggered by zone related logic.
 *
 * However, keep in mind this does not mean `zone.js` is not present.
 * The environment could still run in ZoneFull mode even if Angular turned it off.
 * Consider the situation of a Angular element configured for ZoneLess
 * environments is used in an Angular application relining on the zone mechanism.
 *
 * @param instance {Class Instance} - The instance to check for constructor name of `NgZone`.
 * @return {boolean} - true if instance is of type `NgZone`.
 *
 */
function isNgZone(instance) {
    const cachedValue = zoneDetectionCache.get(instance);
    if (cachedValue !== undefined) {
        return cachedValue;
    }
    let calledApply = false;
    function fn() {
    }
    fn.apply = () => (calledApply = true);
    instance.runOutsideAngular(fn);
    zoneDetectionCache.set(instance, calledApply);
    return calledApply;
}
/**
 * isNoopNgZone
 *
 *@description
 *
 * This function takes any instance of a class and checks
 * if the constructor name is equal to `NoopNgZone`.
 *
 * For more detailed information read the description of [isNgZone](#isngzone).
 *
 * @param instance {Class Instance} - The instance to check for constructor name of `NoopNgZone`.
 * @return {boolean} - true if instance is of type `NoopNgZone`.
 *
 */
function isNoopNgZone(instance) {
    return !isNgZone(instance);
}

/** A shared promise instance to cause a delay of one microtask */
let resolvedPromise = null;
function getUnpatchedResolvedPromise() {
    resolvedPromise = resolvedPromise || getZoneUnPatchedApi('Promise').resolve();
    return resolvedPromise;
}

const coalescingManager = createCoalesceManager();
const ɵ0 = (ctx) => ({
    numCoalescingSubscribers: 0
});
const coalescingContextPropertiesMap = createPropertiesWeakMap(ɵ0);
/**
 * @describe createCoalesceManager
 *
 * returns a
 * Maintains a weak map of component references ans flags
 * them if the coalescing process is already started for them.
 *
 * Used in render aware internally.
 */
function createCoalesceManager() {
    return {
        remove: removeWork,
        add: addWork,
        isCoalescing
    };
    // Increments the number of subscriptions in a scope e.g. a class instance
    function removeWork(scope) {
        const numCoalescingSubscribers = coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers - 1;
        coalescingContextPropertiesMap.setProps(scope, {
            numCoalescingSubscribers: numCoalescingSubscribers >= 0 ? numCoalescingSubscribers : 0
        });
    }
    // Decrements the number of subscriptions in a scope e.g. a class instance
    function addWork(scope) {
        const numCoalescingSubscribers = coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers +
            1;
        coalescingContextPropertiesMap.setProps(scope, {
            numCoalescingSubscribers
        });
    }
    // Checks if anybody else is already coalescing atm
    function isCoalescing(scope) {
        return (coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers >
            0);
    }
}

/**
 *
 * @description
 *
 * This function takes an elem and event and re-applies the listeners from the passed event to the
 * passed element with the zone un-patched version of it.
 *
 * @param elem {HTMLElement} - The elem to re-apply the listeners to.
 * @param event {string} - The name of the event from which to re-apply the listeners.
 *
 * @returns void
 */
function unpatchEventListener(element, event) {
    // `EventTarget` is patched only in the browser environment, thus
    // running this code on the server-side will throw an exception:
    // `TypeError: element.eventListeners is not a function`.
    if (typeof element.eventListeners !== 'function') {
        return;
    }
    const eventListeners = element.eventListeners(event);
    // Return if no event listeners are present
    if (!eventListeners) {
        return;
    }
    const addEventListener = getZoneUnPatchedApi('addEventListener', element).bind(element);
    eventListeners.forEach((listener) => {
        // Remove and reapply listeners with patched API
        // @TODO use (elem as any).removeAllListeners?(eventName?: string): void;
        element.removeEventListener(event, listener);
        // Reapply listeners with un-patched API
        addEventListener(event, listener);
    });
}
/**
 * @Directive UnpatchEventsDirective
 *
 * @description
 *
 * The `unpatch` directive helps in partially migrating to zone-less apps as well as getting rid
 * of unnecessary renderings through zones `addEventListener` patches.
 * It can be used on any element you apply event bindings.
 *
 * The current way of binding events to the DOM is to use output bindings:
 *  ```html
 * <button (click)="doStuff($event)">click me</button>
 * ```
 *
 * The problem is that every event registered over `()` syntax, e.g. `(click)`
 * marks the component and all its ancestors as dirty and re-renders the whole component tree.
 * This is because zone.js patches the native browser API and whenever one of the patched APIs is used it re-renders.
 *
 * So even if your button is not related to a change that needs a re-render the app will re-render completely.
 * This leads to bad performance. This is especially helpful if you work with frequently fired events like 'mousemove'
 *
 * `unpatch` directive solves that problem.
 *
 * Included Features:
 *  - by default un-patch all registered listeners of the host it is applied on
 *  - un-patch only a specified set of registered event listeners
 *  - works zone independent (it directly checks the widow for patched APIs and un-patches them without the use of `runOutsideZone` which brings more performance)
 *  - Not interfering with any logic executed by the registered callback
 *
 * @usageNotes
 *
 * The `unpatch` directive can be used like shown here:
 * ```html
 * <button [unpatch] (click)="triggerSomeMethod($event)">click me</button>
 * <button [unpatch]="['mousemove']" (mousemove)="doStuff2($event)" (click)="doStuff($event)">click me</button>
 * ```
 *
 * @publicApi
 */
// tslint:disable-next-line:directive-selector
class UnpatchEventsDirective {
    constructor(el) {
        this.el = el;
        this.subscription = new Subscription();
        this.events$ = new BehaviorSubject(zonePatchedEvents);
    }
    reapplyEventListenersZoneUnPatched(events) {
        events.forEach((ev) => {
            unpatchEventListener(this.el.nativeElement, ev);
        });
    }
    ngOnChanges({ events }) {
        if (events && Array.isArray(this.events)) {
            this.events$.next(this.events);
        }
    }
    ngAfterViewInit() {
        this.subscription = this.events$.subscribe((eventList) => {
            this.reapplyEventListenersZoneUnPatched(eventList);
        });
    }
    ngOnDestroy() {
        this.subscription.unsubscribe();
    }
}
UnpatchEventsDirective.decorators = [
    { type: Directive, args: [{ selector: '[unpatch]' },] }
];
UnpatchEventsDirective.ctorParameters = () => [
    { type: ElementRef }
];
UnpatchEventsDirective.propDecorators = {
    events: [{ type: Input, args: ['unpatch',] }]
};

const DECLARATIONS = [UnpatchEventsDirective];
class UnpatchEventsModule {
}
UnpatchEventsModule.decorators = [
    { type: NgModule, args: [{
                declarations: DECLARATIONS,
                exports: DECLARATIONS
            },] }
];

function rxMaterialize() {
    return o$ => o$.pipe(materialize(), tap(({ kind, error }) => {
        if (kind === 'E') {
            console.error(error);
        }
    }), map(({ value, hasValue, error, kind }) => ({
        value,
        hasValue,
        error,
        kind: notificationKindToRxNotificationKind(kind)
    })));
}
function notificationKindToRxNotificationKind(kind) {
    switch (kind) {
        case 'C':
            return 'rxComplete';
        case 'E':
            return 'rxError';
        case 'N':
        default:
            return 'rxNext';
    }
}

/**
 * RenderAware
 *
 * @description
 * This function returns an object that holds all the shared logic for the push pipe and the let directive
 * responsible for change detection
 * If you extend this class you need to implement how the update of the rendered value happens.
 * Also custom behaviour is something you need to implement in the extending class
 */
function createRenderAware(cfg) {
    const strategyName$ = new ReplaySubject(1);
    let currentStrategy;
    const strategy$ = strategyName$.pipe(distinctUntilChanged(), switchMap((stringOrObservable) => typeof stringOrObservable === 'string'
        ? of(stringOrObservable)
        : stringOrObservable), map((strategy) => {
        const s = cfg.strategies[strategy];
        if (!!s) {
            return s;
        }
        throw new Error(`Strategy ${strategy} does not exist.`);
    }), tap((s) => (currentStrategy = s)), 
    // do not repeat the steps before for each subscriber
    shareReplay({ bufferSize: 1, refCount: true }));
    const observablesFromTemplate$ = new ReplaySubject(1);
    const valuesFromTemplate$ = observablesFromTemplate$.pipe(distinctUntilChanged());
    let firstTemplateObservableChange = true;
    const renderingEffect$ = valuesFromTemplate$.pipe(
    // handle null | undefined assignment and new Observable reset
    map((observable$) => {
        if (observable$ === null) {
            return of(null);
        }
        if (!firstTemplateObservableChange) {
            cfg.templateObserver.suspense();
            if (observable$ === undefined) {
                return of(undefined);
            }
        }
        firstTemplateObservableChange = false;
        return observable$;
    }), 
    // forward only observable values
    filter((o$) => o$ !== undefined), distinctUntilChanged(), switchMap((o$) => o$
        // Added behavior will get applied to the observable in `renderWithLatestStrategy`
        .pipe(
    // Forward only distinct values
    distinctUntilChanged(), 
    // Update completion, error and next
    tap(cfg.templateObserver), renderWithLatestStrategy(strategy$))), publish());
    return {
        nextPotentialObservable(value) {
            observablesFromTemplate$.next(value);
        },
        nextStrategy(nextConfig) {
            strategyName$.next(nextConfig);
        },
        rendered$: renderingEffect$,
        activeStrategy$: strategy$,
        subscribe() {
            return new Subscription().add(renderingEffect$.connect());
        }
    };
}
function renderWithLatestStrategy(strategyChanges$) {
    const suspenseNotification = {
        kind: 'rxSuspense',
        value: undefined,
        hasValue: false,
        error: undefined,
    };
    return (o$) => {
        return o$.pipe(rxMaterialize(), withLatestFrom(strategyChanges$), 
        // always use latest strategy on value change
        switchMap(([renderValue, strategy]) => concat(of(renderValue), NEVER).pipe(strategy.rxScheduleCD)), startWith(suspenseNotification));
    };
}

/** @internal */
const RxLetTemplateNames = Object.assign(Object.assign({}, RxBaseTemplateNames), { next: 'nextTpl' });
/**
 * @Directive LetDirective
 *
 * @description
 *
 * The `*rxLet` directive serves a convenient way of binding observables to a view context. Furthermore, it helps
 * you structure view-related models into view context scope (DOM element's scope).
 *
 * Under the hood, it leverages a `RenderStrategy` which in turn takes care of optimizing the change detection
 * of your component or embedded view. The `LetDirective` will render its template and manage change detection after it
 *   got an initial value. So if the incoming `Observable` emits its value lazily (e.g. data coming from `Http`), your
 *   template will be rendered lazily as well. This can very positively impact the initial render performance of your
 *   application.
 *
 *
 * ### Problems with `async` and `*ngIf`
 *
 * In Angular, a way of binding an observable to the view could look like that:
 * ```html
 * <ng-container *ngIf="observableNumber$ | async as n">
 *   <app-number [number]="n"></app-number>
 *   <app-number-special [number]="n"></app-number-special>
 * </ng-container>
 * ```
 *
 * The problem is that `*ngIf` interferes with rendering and in case of a `0` (a falsy value) the component
 * would be hidden. This issue doesn't concern the `LetDirective`.
 *
 * The `AsyncPipe` relies on the Zone to be present - it doesn't really trigger change detection by itself.
 * It marks the component and its children as dirty waiting for the Zone to trigger change detection. So, in case
 * you want to create a zone-less application, the `AsyncPipe` won't work as desired. `LetDirective` comes
 * with its own strategies to manage change detection every time a new notification is sent from
 * the bound Observable.
 *
 *
 * ### Features of `*rxLet`
 *
 * Included features for `*rxLet`:
 * - binding is always present. (see "Problems with `async` and `*ngIf`" section below)
 * - it takes away the multiple usages of the `async` or `push` pipe
 * - a unified/structured way of handling null and undefined
 * - triggers change-detection differently if `zone.js` is present or not (`ChangeDetectorRef.detectChanges` or
 *   `ChangeDetectorRef.markForCheck`)
 * - triggers change-detection differently if ViewEngine or Ivy is present (`ChangeDetectorRef.detectChanges` or
 *   `ɵdetectChanges`)
 * - distinct same values in a row (`distinctUntilChanged` operator),
 * - display custom templates for different observable notifications (rxSuspense, rxNext, rxError, rxComplete)
 * - notify about after changes got rendered to the template (RenderCallback)
 *
 *
 * ### Binding an Observable and using the view context
 *
 * The `*rxLet` directive takes over several things and makes it more convenient and save to work with streams in the
 * template:
 *
 * ```html
 * <ng-container *rxLet="observableNumber$; let n">
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 *
 * <ng-container *rxLet="observableNumber$ as n">
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 * ```
 *
 * In addition to that it provides us information from the whole observable context.
 * We can track the observables:
 * - next value
 * - error occurrence
 * - complete occurrence
 *
 * ```html
 * <ng-container *rxLet="observableNumber$; let n; let e = $rxError, let c = $rxComplete">
 *   <app-number [number]="n" *ngIf="!e && !c"></app-number>
 *   <ng-container *ngIf="e">
 *     There is an error: {{ e }}
 *   </ng-container>
 *   <ng-container *ngIf="c">
 *     Observable completed: {{ c }}
 *   </ng-container>
 * </ng-container>
 * ```
 *
 *
 * ### Using the template-binding
 *
 * You can also use template anchors and display template's content for different observable states:
 * - on complete
 * - on error
 * - on suspense - before the first value is emitted
 *
 * ```html
 * <ng-container
 *   *rxLet="
 *     observableNumber$;
 *     let n;
 *     rxError: error;
 *     rxComplete: complete;
 *     rxSuspense: suspense;
 *   "
 * >
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 * <ng-template #error>ERROR</ng-template>
 * <ng-template #complete>COMPLETE</ng-template>
 * <ng-template #suspense>SUSPENSE</ng-template>
 * ```
 *
 * Internally, `*rxLet` is using a simple "view memoization" - it caches all anchored template references and re-uses
 * them whenever the observable notification (next/error/complete) is sent. Then, it only updates the context
 * (e.g. a value from the observable) in the view.
 *
 *
 * @docsCategory LetDirective
 * @docsPage LetDirective
 * @publicApi
 */
class LetDirective {
    // TODO: enable when tested and documented properly
    /* @Input('rxLetShowComplete')
     set showComplete(trigger$: Observable<any>) {
     this.triggerHandler.next(
     trigger$.pipe(mapTo(toRxCompleteNotification() as any))
     );
     }
  
     @Input('rxLetShowError')
     set showError(error$: Observable<any>) {
     this.triggerHandler.next(error$.pipe(map(toRxErrorNotification as any)));
     }
  
     @Input('rxLetShowSuspense')
     set showSuspense(trigger$: Observable<any>) {
     this.triggerHandler.next(
     trigger$.pipe(map(toRxSuspenseNotification as any))
     );
     }*/
    constructor(strategyProvider, cdRef, eRef, ngZone, nextTemplateRef, viewContainerRef) {
        this.strategyProvider = strategyProvider;
        this.cdRef = cdRef;
        this.eRef = eRef;
        this.ngZone = ngZone;
        this.nextTemplateRef = nextTemplateRef;
        this.viewContainerRef = viewContainerRef;
        this.renderParent = true;
        this.patchZone = this.strategyProvider.config.patchZone;
        /** @internal */
        this.observablesHandler = templateNotifier(() => !!this.rxSuspense);
        /** @internal */
        this.strategyHandler = hotFlatten(() => new ReplaySubject(1));
        /** @internal */
        this.triggerHandler = hotFlatten(() => new Subject(), mergeAll());
        /** @internal */
        this.subscription = new Subscription();
        /** @internal */
        this.rendered$ = new Subject();
        this.rendered = defer(() => this.rendered$.pipe());
    }
    /**
     * @description
     * The Observable to be bound to the context of a template.
     *
     * @example
     * <ng-container *rxLet="hero$; let hero">
     *   <app-hero [hero]="hero"></app-hero>
     * </ng-container>
     *
     * @param potentialObservable
     */
    set rxLet(potentialObservable) {
        this.observablesHandler.next(potentialObservable);
    }
    /**
     * @description
     * The rendering strategy to be used when rendering with the reactive context within a template.
     * Use it to dynamically manage your rendering strategy. You can switch the strategies
     * imperatively (with a string) or by bounding an Observable.
     * The default strategy is `'local'`.
     *
     * @example
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *     <ng-container *rxLet="hero$; let hero; strategy: strategy">
     *       <app-hero [hero]="hero"></app-hero>
     *     </ng-container>
     *   `
     * })
     * export class AppComponent {
     *   strategy = 'local';
     * }
     *
     * // OR
     *
     * \@Component({
     *   selector: 'app-root',
     *   template: `
     *     <ng-container *rxLet="hero$; let hero; strategy: strategy$">
     *       <app-hero [hero]="hero"></app-hero>
     *     </ng-container>
     *   `
     * })
     * export class AppComponent {
     *   strategy$ = new BehaviorSubject('local');
     * }
     *
     * @param strategy
     * @see {@link strategies}
     */
    set strategy(strategyName) {
        this.strategyHandler.next(strategyName);
    }
    set renderCallback(callback) {
        this._renderObserver = callback;
    }
    /** @internal */
    static ngTemplateContextGuard(dir, ctx) {
        return true;
    }
    /** @internal */
    ngOnInit() {
        this.subscription.add(this.templateManager
            .render(this.observablesHandler.values$)
            .subscribe((n) => {
            var _a;
            this.rendered$.next(n);
            (_a = this._renderObserver) === null || _a === void 0 ? void 0 : _a.next(n);
        }));
    }
    /** @internal */
    ngOnChanges(changes) {
        if (!this.templateManager) {
            this._createTemplateManager();
        }
        if (changes.rxComplete) {
            this.templateManager.addTemplateRef(RxLetTemplateNames.complete, this.rxComplete);
        }
        if (changes.rxSuspense) {
            this.templateManager.addTemplateRef(RxLetTemplateNames.suspense, this.rxSuspense);
        }
        if (changes.rxError) {
            this.templateManager.addTemplateRef(RxLetTemplateNames.error, this.rxError);
        }
    }
    /** @internal */
    ngOnDestroy() {
        this.subscription.unsubscribe();
    }
    /** @internal */
    _createTemplateManager() {
        this.templateManager = createTemplateManager({
            templateSettings: {
                viewContainerRef: this.viewContainerRef,
                createViewContext,
                updateViewContext,
                customContext: (rxLet) => ({ rxLet }),
                patchZone: this.patchZone ? this.ngZone : false,
            },
            renderSettings: {
                cdRef: this.cdRef,
                eRef: this.eRef,
                parent: coerceBooleanProperty(this.renderParent),
                patchZone: this.patchZone ? this.ngZone : false,
                defaultStrategyName: this.strategyProvider.primaryStrategy,
                strategies: this.strategyProvider.strategies,
            },
            notificationToTemplateName: {
                [RxNotificationKind.suspense]: () => this.rxSuspense
                    ? RxLetTemplateNames.suspense
                    : RxLetTemplateNames.next,
                [RxNotificationKind.next]: () => RxLetTemplateNames.next,
                [RxNotificationKind.error]: () => this.rxError ? RxLetTemplateNames.error : RxLetTemplateNames.next,
                [RxNotificationKind.complete]: () => this.rxComplete
                    ? RxLetTemplateNames.complete
                    : RxLetTemplateNames.next,
            },
            templateTrigger$: this.triggerHandler.values$,
        });
        this.templateManager.addTemplateRef(RxLetTemplateNames.next, this.nextTemplateRef);
        this.templateManager.nextStrategy(this.strategyHandler.values$);
    }
}
LetDirective.decorators = [
    { type: Directive, args: [{
                selector: '[rxLet]',
                providers: [],
            },] }
];
LetDirective.ctorParameters = () => [
    { type: RxStrategyProvider },
    { type: ChangeDetectorRef },
    { type: ElementRef },
    { type: NgZone },
    { type: TemplateRef },
    { type: ViewContainerRef }
];
LetDirective.propDecorators = {
    rxLet: [{ type: Input }],
    strategy: [{ type: Input, args: ['rxLetStrategy',] }],
    rxComplete: [{ type: Input, args: ['rxLetRxComplete',] }],
    rxError: [{ type: Input, args: ['rxLetRxError',] }],
    rxSuspense: [{ type: Input, args: ['rxLetRxSuspense',] }],
    renderCallback: [{ type: Input, args: ['rxLetRenderCallback',] }],
    renderParent: [{ type: Input, args: ['rxLetParent',] }],
    patchZone: [{ type: Input, args: ['rxLetPatchZone',] }],
    rendered: [{ type: Output }]
};
/** @internal */
function createViewContext(value) {
    return {
        rxLet: value,
        $implicit: value,
        $error: false,
        $complete: false,
        $suspense: false,
    };
}
/** @internal */
function updateViewContext(value, view, context) {
    Object.keys(context).forEach((k) => {
        view.context[k] = context[k];
    });
}

/**
 *
 * @description
 *
 * This function takes an elem and event and re-applies the listeners from the passed event to the
 * passed element with the zone un-patched version of it.
 *
 * @param elem {HTMLElement} - The elem to re-apply the listeners to.
 * @param event {string} - The name of the event from which to re-apply the listeners.
 *
 * @returns void
 */
function unpatchEventListener$1(elem, event) {
    const eventListeners = elem.eventListeners(event);
    // Return if no event listeners are present
    if (!eventListeners) {
        return;
    }
    const addEventListener = getZoneUnPatchedApi('addEventListener', elem).bind(elem);
    eventListeners.forEach((listener) => {
        // Remove and reapply listeners with patched API
        elem.removeEventListener(event, listener);
        // Reapply listeners with un-patched API
        addEventListener(event, listener);
    });
}
function intersectionObserver(options) {
    const subject = new Subject();
    const observer = observerSupported()
        ? new IntersectionObserver((entries) => {
            entries.forEach((entry) => subject.next(entry));
        }, options)
        : null;
    const entries$ = new Observable((subscriber) => {
        subject.subscribe(subscriber);
        return () => {
            if (observer) {
                observer.disconnect();
            }
        };
    });
    return {
        entries$,
        observe: observer.observe,
        unobserve: observer.unobserve,
    };
}
const observerSupported = () => typeof window !== 'undefined'
    ? !!window.IntersectionObserver
    : false;
const ɵ0$1 = observerSupported;
class ViewportPrioDirective {
    constructor(el, strategyProvider, letDirective) {
        this.el = el;
        this.strategyProvider = strategyProvider;
        this.letDirective = letDirective;
        // Note that we're picking only the `intersectionRatio` property
        // since this is the only property that we're intersted in.
        this.entriesSubject = new Subject();
        this.entries$ = this.entriesSubject.pipe(mergeAll());
        this._viewportPrio = 'noop';
        /* @Input('viewport-prio')
        set viewportPrio(prio) {
          if (prio) {
            this._viewportPrio = prio || 'noop';
          }
        }*/
        this.observer = observerSupported()
            ? new IntersectionObserver((entries) => {
                this.entriesSubject.next(entries);
            }, {
                threshold: 0,
            })
            : null;
        this.visibilityEvents$ = this.entries$.pipe(map((entry) => {
            if (entry.intersectionRatio > 0) {
                return 'visible';
            }
            else {
                return 'invisible';
            }
        }));
    }
    ngOnInit() {
        const letStrategyName$ = this.letDirective['strategyHandler'].values$.pipe(filter((name) => name !== this._viewportPrio));
        this.visibilityEvents$
            .pipe(withLatestFrom(letStrategyName$), map(([visibility, strategyName]) => visibility === 'visible' ? strategyName : this._viewportPrio))
            .subscribe((strategyName) => {
            this.letDirective.strategy = strategyName;
            // render actual state on viewport enter
            // @TODO this doesnt catch unsubscribe (cant be cancelled)
            // @TODO: we need to fetch the current template of the letDirective here
            // this.strategyProvider.scheduleCD()
        });
        // If the browser doesn't support the `IntersectionObserver` or we're inside
        // the Node.js environment, then this will throw an exception that property
        // `observe` doesn't exist on `null`.
        if (this.observer !== null) {
            this.observer.observe(this.el.nativeElement);
        }
        else {
            // If we're inside the Node.js environment then this should be
            // rendered (e.g. for SEO purposes), and when running this code in browser
            // it will decide itself to render it or not.
            this.entriesSubject.next([{ intersectionRatio: 1 }]);
        }
    }
    ngOnDestroy() {
        if (this.observer) {
            this.observer.disconnect();
        }
    }
}
ViewportPrioDirective.decorators = [
    { type: Directive, args: [{
                // tslint:disable-next-line:directive-selector
                selector: '[viewport-prio]',
            },] }
];
ViewportPrioDirective.ctorParameters = () => [
    { type: ElementRef },
    { type: RxStrategyProvider },
    { type: LetDirective, decorators: [{ type: Optional }] }
];

const DECLARATIONS$1 = [ViewportPrioDirective];
class ViewportPrioModule {
}
ViewportPrioModule.decorators = [
    { type: NgModule, args: [{
                declarations: DECLARATIONS$1,
                exports: DECLARATIONS$1
            },] }
];

/**
 * @Pipe PushPipe
 *
 * @description
 *
 * The `push` pipe serves as a drop-in replacement for the `async` pipe.
 * It contains intelligent handling of change detection to enable us
 * running in zone-full as well as zone-less mode without any changes to the code.
 *
 * The current way of binding an observable to the view looks like that:
 *  ```html
 *  {{observable$ | async}}
 * <ng-container *ngIf="observable$ | async as o">{{o}}</ng-container>
 * <component [value]="observable$ | async"></component>
 * ```
 *
 * The problem is `async` pipe just marks the component and all its ancestors as dirty.
 * It needs zone.js microtask queue to exhaust until `ApplicationRef.tick` is called to render all dirty marked
 *     components.
 *
 * Heavy dynamic and interactive UIs suffer from zones change detection a lot and can
 * lean to bad performance or even unusable applications, but the `async` pipe does not work in zone-less mode.
 *
 * `push` pipe solves that problem.
 *
 * Included Features:
 *  - Take observables or promises, retrieve their values and render the value to the template
 *  - Handling null and undefined values in a clean unified/structured way
 *  - Triggers change-detection differently if `zone.js` is present or not (`detectChanges` or `markForCheck`)
 *  - Distinct same values in a row to increase performance
 *  - Coalescing of change detection calls to boost performance
 *
 * @usageNotes
 *
 * `push` pipe solves that problem. It can be used like shown here:
 * ```html
 * {{observable$ | push}}
 * <ng-container *ngIf="observable$ | push as o">{{o}}</ng-container>
 * <component [value]="observable$ | push"></component>
 * ```
 *
 * @publicApi
 */
class PushPipe {
    constructor(strategyProvider, ngZone, cdRef) {
        this.strategyProvider = strategyProvider;
        this.ngZone = ngZone;
        /** @internal */
        this.templateObserver = templateNotifier();
        /** @internal */
        this.strategyHandler = strategyHandling(this.strategyProvider.primaryStrategy, this.strategyProvider.strategies);
        const scope = cdRef.context;
        this.subscription = this.templateObserver.values$
            .pipe(filter((n) => n.kind === 'suspense' || n.kind === 'next'), tap((notification) => {
            this.renderedValue = notification.value;
        }), withLatestFrom(this.strategyHandler.strategy$), switchMap(([v, strategy]) => this.strategyProvider.schedule(() => {
            strategy.work(cdRef, scope);
        }, {
            scope,
            strategy: strategy.name,
            patchZone: this.strategyProvider.config.patchZone
                ? ngZone
                : false,
        })))
            .subscribe();
    }
    transform(potentialObservable, config, renderCallback) {
        if (config) {
            this.strategyHandler.next(config);
        }
        this.templateObserver.next(potentialObservable);
        return this.renderedValue;
    }
    /** @internal */
    ngOnDestroy() {
        this.subscription.unsubscribe();
    }
}
PushPipe.decorators = [
    { type: Pipe, args: [{ name: 'push', pure: false },] }
];
PushPipe.ctorParameters = () => [
    { type: RxStrategyProvider },
    { type: NgZone },
    { type: ChangeDetectorRef }
];

const DECLARATIONS$2 = [PushPipe];
/**
 * @description
 * This module exports the PushPipe
 *
 * @example
 *
 * ```typescript
 * @NgModule({
 *  imports: [PushModule],
 *  //...
 * })
 * export class AppModule {}
 * ```
 *
 */
class PushModule {
}
PushModule.decorators = [
    { type: NgModule, args: [{
                declarations: DECLARATIONS$2,
                imports: [],
                exports: DECLARATIONS$2
            },] }
];

const EXPORTED_DECLARATIONS = [LetDirective];
class LetModule {
}
LetModule.decorators = [
    { type: NgModule, args: [{
                declarations: EXPORTED_DECLARATIONS,
                exports: [EXPORTED_DECLARATIONS]
            },] }
];

// EXPERIMENTAL

/**
 * Generated bundle index. Do not edit.
 */

export { LetDirective, LetModule, PushModule, PushPipe, UnpatchEventsDirective, UnpatchEventsModule, ViewportPrioDirective, ViewportPrioModule };
//# sourceMappingURL=rx-angular-template.js.map
